# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
# and published under the Terms of the GNU Affero General Public License.
# Text of the AGPL: http://www.gnu.org/licenses/agpl-3.0-standalone.html
# credits: http://www.opsi.org/credits/

#***************************************************************************
# Subversion:
# $Revision: 40 $
# $Author: oertel $
# $Date: 2015-01-16 18:17:00 +0100 (Fr, 16 Jan 2015) $
#***************************************************************************

# Debian
# https://www.debian.org/releases/stable/amd64/ch05s03.html.de
# https://wiki.debian.org/DebianInstaller/
# https://www.debian.org/releases/stable/amd64/apb.html.de
# preseed files:
# http://www.debian-administration.org/article/394/Automating_new_Debian_installations_with_preseeding

# Ubuntu:
#http://archive.ubuntu.com/ubuntu/dists/trusty-updates/main/installer-amd64/current/images/netboot/ubuntu-installer/amd64/
#https://wiki.ubuntu.com/UbiquityAutomation
#https://help.ubuntu.com/14.04/installation-guide/example-preseed.txt

try:
	import ConfigParser
except:
	import configparser
from OPSI.Util.File import IniFile
from OPSI.Util.File import TextFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from OPSI.Util import blowfishDecrypt
from hashlib import md5
try:
	from twisted.conch.ssh import keys
except:
	pass
try:
	from urlparse import urlparse
except:
	from urllib.parse import urlparse
from os.path import splitext, basename
import platform
import re, xml.dom.minidom, os, gzip
import crypt
from base64 import b64encode
from os import urandom

if not bootimageVersion or (int(bootimageVersion) < 20210618):
	raise Exception(u"This product requires opsi-linux-bootimage version >= 20210618")
try:
	if not backend.backend_info()['opsiVersion'].startswith('4.'):
		raise ValueError('No opsi 4.x backend present')
except Exception:
	raise Exception(u"This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

if isOpsi43:
    show_message(u"Installing Product: %s" % (productId), 10)

opsiVersion = backend.backend_info()['opsiVersion']
source = SCRIPT_PATH

# Do hardware inventory
logger.notice(u"Fetching opsi hw audit configuration")
hwconfig = backend.auditHardware_getConfig()

if isOpsi43:
    show_message(u"Running hardware inventory", 10)
logger.notice(u"Running hardware inventory")
auditHardwareOnHosts = auditHardware(config = hwconfig, hostId = clientId)

logger.notice(u"Sending hardware information to service")
backend.auditHardwareOnHost_setObsolete(clientId)
backend.auditHardwareOnHost_createObjects(auditHardwareOnHosts)

ramSize = execute("grep MemTotal /proc/meminfo | awk '{ print $2 }'")[0]
if int(ramSize) < 4000000:
	raise Exception("Memory is less than 4GB. A network installation requires at least 4GB of RAM")

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsipreparelib.py','/tmp')
exec(open('/tmp/opsipreparelib.py').read())

initOpsiSetupLib()

useGpt = False
inUefi = inUefiMode()
inSecureboot = False
if inUefi:
	try:
		inSecureboot = isSecureboot()
	except NameError:
		isSecureboot = False

	if inSecureboot:
		logger.notice(u"we are running in Secure Boot mode")
		scriptMessageSubject.setMessage(u"we are running in Secure Boot mode")
	else:
		logger.notice(u"we are not running in Secure Boot mode")
		scriptMessageSubject.setMessage(u"we are not running in Secure Boot mode")
	useGpt = True

if useGpt:
	logger.notice(u"Will use GPT")
else:
	logger.notice(u"Will use MBR Partitions")

# create network path
if (depot._url.startswith('smb:') or depot._url.startswith('cifs:')):
	match = re.search('^smb://([^/]+)/([^/]+)(.*)$', depot._url, re.IGNORECASE)
	if not match:
		raise Exception("Bad depot-URL '%s'" % depot._url)
	hn = match.group(1)
	sn = match.group(2)
	pn = match.group(3)
else:
	match = re.search('^webdavs://([^/]+):([^/]+)(.*)$', depot._url, re.IGNORECASE)
	if not match:
		raise Exception("Bad depot-URL '%s'" % depot._url)
	hn = match.group(1)
	sn = match.group(2)
	pn = match.group(3)

depotDir = '/var/lib/opsi/depot'
nfsserver = hn
nfssubdir = 'opsi_nfs_share'
nfsbasedir = '%s/%s' % (depotDir, nfssubdir)
nfsclientconfig =  configStates.get('clientconfig.opsi_nfs_share', [ '' ])[0]
if nfsclientconfig:
	nfsserver = nfsclientconfig.split(":")[0]
	nfsbasedir = nfsclientconfig.split(":")[1]
logger.notice(u"Using as nfsserver: %s" % nfsserver)
logger.notice(u"Using as nfsbasedir: %s" % nfsbasedir)


proxy = productProperties.get('proxy','')

install_opsi_client_agent = False
do_opsi_client = "false"
if (str(productProperties.get('install_opsi-client-agent', 'true')).lower() == 'true'):
	install_opsi_client_agent = True
	do_opsi_client = "true"

install_unattended = True
if (str(productProperties.get('install_unattended', 'true')).lower() == 'false'):
	install_unattended = False

installation_method = productProperties.get('installation_method', 'kexec')
if inSecureboot:
	installation_method = 'reboot'
	logger.notice("We are running in SecureBoot mode, enforcing reboot installation_method")

locale = productProperties.get('locale','de_DE')
language = locale.split('_')[0]
country = locale.split('_')[1]
keymap = productProperties.get('console_keymap', 'de-latin1-nodeadkeys')
try:
	xkbvariant = keymap.split('-')[2]
except:
	xkbvariant = ''
	pass
xkblayout = keymap.split('-')[0]

online_repository = productProperties.get('online_repository', 'http://de.archive.ubuntu.com/ubuntu')

if (str(productProperties.get('use_nonfree', 'true')).lower() == 'true'):
	use_nonfree = "true"
else:
	use_nonfree = "false"

myarchitecture = getArchitecture()
if myarchitecture == 'amd64' and 'opsi' in online_repository:
        online_repository = online_repository + '/amd64'
if myarchitecture == 'i386' and 'opsi' in online_repository:
        online_repository = online_repository + '/i386'

logger.notice(u"splitting online_repository ...")
online_repository_protocol=online_repository.split('://')[0]
online_repo_adr = online_repository.split('://')[1]
online_repository_base=online_repo_adr.split('/')[0]
# get the rest: split has num argument
online_repository_dir='/'+online_repo_adr.split('/',1)[1]

partition_disk = productProperties.get('partition_disk', 'first')
partition_method = productProperties.get('partition_method', 'lvm')
#if inUefi:
#	partition_method = 'efi'
partition_recipe = productProperties.get('partition_recipe', 'atomic')
if partition_recipe == 'customSwap':
	partition_method = 'regular'
	swapSizeProp = productProperties.get('swap_size', '')
	if swapSizeProp:
		removedUnits = removeUnit(swapSizeProp)
		swapSize = int(removedUnits)/1048576
		logger.notice(u"swap size: %s" % swapSize)
	else:
		meminfo = open('/proc/meminfo').read()
		matched = re.search(r'^MemTotal:\s+(\d+)', meminfo)
		if matched:
			memory = int(matched.groups()[0])
			swapSize = int(memory)/1024
			logger.notice(u"memory: %s" % swapSize)


#if inUefi:
#	partition_recipe = 'amd64/efi'
desktop_package = productProperties.get('desktop_package', 'standard')
#language_packs = productProperties.get('language_packs', 'standard')
upgrade_select = productProperties.get('upgrade_select', 'none')


my_broadcast = getBroadcastAddress(ipaddress, netmask)
my_network = getNetworkAddress(ipaddress, netmask)

# Get and encrypt passwords
# user
username = productProperties.get('user_name', 'user')
userpassclear = productProperties.get('user_password', 'linux123')
userpassenc = shadow_password_hash(userpassclear)

# root
rootpassclear = productProperties.get('root_password', 'linux123')
rootpassenc = shadow_password_hash(rootpassclear)

if (proxy == ''):
	do_proxy = 'false'
	proxy_addr = ''
else:
	do_proxy = 'true'
	proxy_addr = proxy
	if proxy_addr.startswith('https:'):
		execute('export HTTPS_PROXY="%s"' % proxy_addr)
	else:
		execute('export HTTP_PROXY="%s"' % proxy_addr)

opsi_client_agent_install_script = ''

# get machine architecture
scriptMessageSubject.setMessage(u"Installing 64 Bit system....")
logger.notice(u"Installing 64 Bit system....")
myarchitectureBits = '64'

if isOpsi43:
    show_message("Preparing disk", 20)
scriptMessageSubject.setMessage("Create temporary place on disk.... ")
logger.notice(u"Create temporary place on disk.... ")
#prepareDisk()
target = '/mnt/hd'
systemPartitionNumber = 1
if inUefi:
	systemPartitionNumber = 3
	if installation_method == 'reboot':
		grubPartitionNumber = 3
		systemPartitionNumber +=1
		grubPartitionSizeM = 2000
		grubPartitionLabel = 'INSTALLER'
systemPartitionSizeM = 6000
systemPartitionLabel  = 'LINUX'
blockAlignment = True

logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
lvmClearVolumeGroups()
execute('sfdisk -l')

# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
# use first disk
disk = disks[0]
if partition_disk == "first":
	partition_disk = disk.device

if useGpt:
	logger.notice(u"Clear partition table of %s" % (disk.device))
	execute('/usr/sbin/sgdisk --zap-all %s' % (disk.device))
	execute('/usr/sbin/sgdisk --clear -g -p %s' % (disk.device))
	if inUefi:
		currentbootentry = uefiGetActBootEntry()
		if currentbootentry:
			uefiNetbootLabel = currentbootentry['label']
			logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
			backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label")
			backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
		else:
			logger.notice('No usable current (net) boot entry found')

		deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
		deleteAllEntriesFromEFIBootmgr('Windows Boot Manager')
		deleteAllEntriesFromEFIBootmgr('ubuntu')
		deleteAllEntriesFromEFIBootmgr('debian')
		deleteAllEntriesFromEFIBootmgr('installer')
		deleteAllEntriesFromEFIBootmgr('opsibootimage')
		logger.notice(u"Create EFI system partition on %s1" % (disk.device))
		execute('/usr/sbin/sgdisk -g --new 1::100M -t 1:ef00 -c 1:\"EFI system partition\" --partition-guid=C12A7328-F81F-11D2-BA4B-00A0C93EC93B -p %s' % (disk.device))
		execute('/bin/sleep 2')
		execute('/usr/sbin/sgdisk -p %s' % (disk.device))
		logger.notice(u"Make file system on EFI system partition on %s1" % (disk.device))
		normalDiskDevices = ['/dev/sda', '/dev/vda', '/dev/sdb']
		if any(normalDisk in disk.device for normalDisk in normalDiskDevices):
			execute('/sbin/mkfs.vfat %s1' % (disk.device))
		else:
			execute('/sbin/mkfs.vfat %sp1' % (disk.device))
		logger.notice(u"Create MS reserved partition on %s2" % (disk.device))
		execute('/usr/sbin/sgdisk -g --new 2::+128M -t 2:0c01 -c 2:\"MS reserved\" --partition-guid=E3C9E316-0B5C-4DB8-817D-F92DF00215AE -p %s' % (disk.device))
else:
	disk.deletePartitionTable()

depotAddress = depotUrl.split('/')[2].replace(':4447','')
try:
	isofilename = execute(f"wget -d -r -np -N --spider -e robots=off --no-check-certificate https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/iso/ 2>&1 | grep ' -> ' | grep -Ev '\/\?C=' | sed 's/.* -> //' | grep '.*\.iso$' | rev | cut -d '/' -f 1 | rev")[0]
except IndexError:
	raise Exception('No iso found in iso directory')
logger.notice(u"Will use iso file: %s " % isofilename)

custom_postinstd = False
wget_exitcode= execute(f'wget -q --spider --no-check-certificate --http-user {fqdn} --http-passwd {pckey} https://{depotAddress}:4447/depot/{productId}/custom/postinst.d/ && echo "200" || echo "1"')
if wget_exitcode[0] == "200":
	custom_postinstd = f'in-target wget -r -nH --no-check-certificate --cut-dirs=3 --no-parent --http-user {fqdn} --http-passwd {pckey} --reject="index.html*" https://{depotAddress}:4447/depot/{productId}/custom/postinst.d/ -P /var/tmp/opsi-linux-client-agent/files; \ '

mint_kernel_append_parameter = productProperties.get('mint_kernel_append_parameter', '')

if installation_method == 'reboot':
	if inUefi:
		grubTarget = '/mnt/grub/'
		checkCreateDir(grubTarget)
		execute('/usr/sbin/sgdisk --info=%d %s' % (grubPartitionNumber, disk.device))
		createPartitionEx("installer", grubPartitionNumber, grubPartitionLabel, 0, grubPartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "63" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "fat32", gptFileSystem="0700")
		execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 --attributes %d:clear:0 -p %s' % (grubPartitionNumber, grubPartitionNumber, grubPartitionNumber, disk.device))
		testPartitionEx(disk.device, grubPartitionNumber, useGpt)
		checkClearPartition("none", disk.device, grubPartitionNumber)
		createFilesystemEx("bootimage", grubPartitionNumber, "bootimage", "fat32", disk.device , useGpt)
		checkMount(grubTarget, disk.device, grubPartitionNumber, "vfat")
		logger.notice("Copying files to partition %d" % grubPartitionNumber)
		depot.copy('/%s/installfiles64/*' % (productId), grubTarget)
		execute('sed -i s_/linux_/vmlinuz_g %s/grub/grub.cfg' % (grubTarget))
		append_line = f'{mint_kernel_append_parameter} locale={locale} keymap={keymap} console-keymaps-at/keymap={keymap} ip=dhcp boot=casper root=dev/ram0 ramdisk_size=3500000 url=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/iso/{isofilename} initrd=mint/initrd.lz nosplash automatic-ubiquity --'
		newLines = []
		with open('%s/grub/grub.cfg' % (grubTarget), 'r+') as grubConfig:
			for line in grubConfig:
				if line.strip().startswith('linux'):
					newLines.append('%s %s\n' % (line.rstrip(), append_line.rstrip()))
					continue
				newLines.append(line)

		with open('%s/grub/grub.cfg' % (grubTarget), 'w') as grubConfig:
			grubConfig.writelines(newLines)
		execute('/bin/efibootmgr -v')
	else:
		logger.warning(u"reboot installation method only works using UEFI. Reverting back to kexec method")
		installation_method = "kexec"

fdiskexitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
createPartitionEx("system", systemPartitionNumber, systemPartitionLabel, 0, systemPartitionSizeM, gptAutoAppend = True, boot = False, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ext4", gptFileSystem="8300")
createFilesystemEx("system", systemPartitionNumber, systemPartitionLabel, "ext4", disk.device, useGpt)
systemPartitionDevicePath = getPartitionDevicePath(disk.device, systemPartitionNumber)

# Partition table deleted => delete all netboot product states
deletePocs = []
for poc in backend.productOnClient_getObjects(clientId = clientId):
	if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
	   ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
		deletePocs.append(poc)
backend.productOnClient_deleteObjects(deletePocs)

if isOpsi43:
    show_message("Mounting partitions and copying data", 40)
# Mount temporary partition
checkCreateDir(target)
checkClearMountpoint(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkClearPartition(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkMount(target, disk.device, systemPartitionNumber, "ext4", systemPartitionDevicePath)
checkCreateDir(target+'/tmp')

scriptMessageSubject.setMessage("Getting kernel and initrd.... ")
logger.notice(u"Getting kernel and initrd.... ")

execute(f'mkdir {target}/iso')
depot.copy(f'/{productId}/iso/{isofilename}', f'{target}/iso/')
execute('mkdir -p /mnt/iso/')
execute(f'mount {target}/iso/{isofilename} /mnt/iso/')
execute(f'cp /mnt/iso/casper/vmlinuz {target}/tmp/')
execute(f'cp /mnt/iso/casper/initrd.lz {target}/tmp/initrd.raw')
block1Output = execute(f"dd if={target}/tmp/initrd.raw skip=0 | cpio -it")
for line in block1Output:
	match = re.search("^[-+]?[0-9]+", line)
	if match:
		block1 = forceInt(match.group(0))
block2Output = execute(f"dd if={target}/tmp/initrd.raw skip={block1} | cpio -it")
for line in block2Output:
	match = re.search("^[-+]?[0-9]+", line)
	if match:
		block2 = forceInt(match.group(0))
blocksum = int(block1) + int(block2)
if 'mint22' in productId.lower():
	execute(f'dd if={target}/tmp/initrd.raw skip={blocksum} of={target}/tmp/initrd.firmware')
	block3Output = execute(f"dd if={target}/tmp/initrd.raw skip={blocksum} | cpio -it")
	for line in block3Output:
		match = re.search("^[-+]?[0-9]+", line)
		if match:
			block3 = forceInt(match.group(0))
			blocksum = int(block1) + int(block2) + int(block3)
			execute(f'dd if={target}/tmp/initrd.raw skip={blocksum} of={target}/tmp/initrd.zst')
			execute(f'unzstd -f {target}/tmp/initrd.zst')
elif 'mint21' in productId.lower():
	execute(f'dd if={target}/tmp/initrd.raw skip={blocksum} of={target}/tmp/initrd.zst')
	execute(f'unzstd -f {target}/tmp/initrd.zst')
else:
	execute(f'dd if={target}/tmp/initrd.raw skip={blocksum} of={target}/tmp/initrd.lz4')
	execute(f'unlz4 -d {target}/tmp/initrd.lz4 {target}/tmp/initrd')

scriptMessageSubject.setMessage("Uncompressing initrd.... ")
logger.notice(u"Uncompressing initrd.... ")
try:
	execute("ls -l %s/tmp/initrd*" % (target))
except:
	pass

scriptMessageSubject.setMessage("Patching initrd.... ")
logger.notice(u"Mounting initrd ....")
execute("mkdir -p %s/tmp/mnt" % (target))
#execute("cd %s/tmp/mnt ; unlz4 ../initrd.lz initrd" % (target))
execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd" % (target))
if 'mint22' in productId.lower():
	execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd.firmware" % (target))
#sys.exit(1)
logger.notice(u"Copy profile for patch....")
preseed = productProperties.get('preseed', 'auto')
if preseed == 'auto':
        logger.notice(u"Default preseed is used")
        depot.copy(SCRIPT_PATH+'/%s.cfg' % (productId), target+'/tmp/installation_profile')
else:
        logger.notice(u"Custom installation profile (%s) found and will be used" % preseed)
        depot.copy(SCRIPT_PATH+'/custom/%s' % preseed, target+'/tmp/installation_profile')

logger.notice(u"Copy /etc/default/keyboard for patch....")
depot.copy(SCRIPT_PATH+'/keyboard', target+'/tmp/keyboard')

# create network path to install iso
if (depot._url.startswith('smb:') or depot._url.startswith('cifs:')):
	match = re.search('^smb://([^/]+)/([^/]+)(.*)$', depot._url, re.IGNORECASE)
	if not match:
		raise Exception("Bad depot-URL '%s'" % depot._url)
	hn = match.group(1)
	sn = match.group(2)
	pn = match.group(3)
else:
	match = re.search('^webdavs://([^/]+):([^/]+)(.*)$', depot._url, re.IGNORECASE)
	if not match:
		raise Exception("Bad depot-URL '%s'" % depot._url)
	hn = match.group(1)
	sn = match.group(2)
	pn = match.group(3)

localUbuntuRepo = False
if (('ubuntu' in productId.lower()) and (online_repository != 'http://de.archive.ubuntu.com/ubuntu')):
        localUbuntuRepo = True

logger.notice(u"Prepare patch file....")
opsipatchscript='/sbin/opsipatchscript.sh'
with open('%s/tmp/mnt/%s' % (target,opsipatchscript), "a") as myfile:
	myfile.write('#! /bin/sh\n')
	myfile.write('set -x\n')
	myfile.write('chroot /target install-keymap %s\n' % keymap)
	myfile.write('cp /opsitmp/keyboard /target/etc/default/keyboard\n')
	#myfile.write('chroot /target cp -a /etc/default/keyboard /tmp/\n')
	#myfile.write('chroot /target cat /tmp/keyboard | sed s"/XKBLAYOUT=\".*/XKBLAYOUT=\"%s\"/" > /etc/default/keyboard\n' % language)
try:
	with open('%s/tmp/mnt/%s' % (target,opsipatchscript), "a") as myfile:
		myfile.write('in-target export DEBIAN_FRONTEND=noninteractive\n')
		myfile.write('in-target apt-get update\n')
		myfile.write('in-target apt-get install console-data xterm librsync1 pkg-config -q --yes\n')
		myfile.write('in-target apt-get install %s -q --yes\n' % desktop_package)
		myfile.write('in-target install-keymap %s\n' % keymap)
		myfile.write('in-target dpkg-reconfigure -f noninteractive console-setup\n')
except requests.exceptions.RequestException as e:
	pass

execute('chmod u+x %s/tmp/mnt/%s\n' % (target,opsipatchscript))
logger.notice('Here comes the %s/tmp/mnt/%s' % (target,opsipatchscript))
execute('cat %s/tmp/mnt/%s' % (target,opsipatchscript))

#
logger.notice(u"Copy SYSCONF_INI....")
# Copy sysconf.ini to opsi-client-agent config dir
copy(SYSCONF_INI , target +'/tmp/')

if partition_recipe == 'customSwap':
	diskSizeM = (int(disk.size)/(1024*1024))
	rootSize = diskSizeM-swapSize-500

for device, lease in leasesFiles.items():
    networkdevice = device

if isOpsi43:
    show_message("patching files", 90)
logger.notice(u"Prepare patch input....")
with open(PATCHA_IN, "a") as myfile:
	myfile.write('networkaddress=%s\n' % my_network)
	myfile.write('broadcastaddress=%s\n' % my_broadcast)
	myfile.write('nameserveraddress=%s\n' % nameserver)
	myfile.write('networkdevice=%s\n' % networkdevice)
	myfile.write('do_opsi_client=%s\n' % do_opsi_client)
	myfile.write('username=%s\n' % username)
	myfile.write('userpassenc=%s\n' % userpassenc)
	myfile.write('rootpassenc=%s\n' % rootpassenc)
	myfile.write('do_proxy=%s\n' % do_proxy)
	myfile.write('proxy_addr=%s\n' % proxy_addr)
	myfile.write('online_repository=%s\n' % online_repository)
	myfile.write('online_repository_protocol=%s\n' % online_repository_protocol)
	myfile.write('online_repository_base=%s\n' % online_repository_base)
	myfile.write('online_repository_dir=%s\n' % online_repository_dir)
	myfile.write('opsi_client_agent_install_script=%s\n' % opsi_client_agent_install_script)
	myfile.write('partition_disk=%s\n' % partition_disk)
	myfile.write('partition_method=%s\n' % partition_method)
	myfile.write('partition_recipe=%s\n' % partition_recipe)
	myfile.write('desktop_package=%s\n' % desktop_package)
	myfile.write('language=%s\n' % language)
	myfile.write('country=%s\n' % country)
	myfile.write('opsipatchscript=%s\n' % opsipatchscript)
	myfile.write('xkbvariant=%s\n' % xkbvariant)
	myfile.write('xkblayout=%s\n' % xkblayout)
	myfile.write('nfsserver=%s\n' % nfsserver)
	myfile.write('nfsdir=%s/%s/%s\n' % (nfsbasedir, productId, myarchitectureBits))
	myfile.write('nfsbasedir=%s\n' % (nfsbasedir))
	myfile.write('upgrade_select=%s\n' % (upgrade_select))
	myfile.write('productid=%s\n' % productId)
	myfile.write('service_ip_address=%s\n' % (nextserver.replace("/rpc", "").rstrip("/")))
	myfile.write('depotAddress=%s\n' % (depotUrl.split('/')[2].replace(':4447','')))
	myfile.write('use_nonfree=%s\n' % use_nonfree)
	if custom_postinstd:
		myfile.write('custom_postinstd=%s\n' % custom_postinstd)
if partition_recipe == 'customSwap':
	with open(PATCHA_IN, "a") as myfile:
		myfile.write('rootSize=%s\n' % (rootSize))
		myfile.write('swapSize=%s\n' % (swapSize))
	logger.notice(u"rootSize: %s, swapSize: %s, method=%s" % (rootSize, swapSize, partition_method))

logger.notice(u"Here comes the PATCHA_IN (%s):" % PATCHA_IN)
execute('cat %s' % (PATCHA_IN))

logger.notice(u"Patching installation_profile")

if not custom_postinstd:
	execute('sed -i "/custom_postinstd/d" %s/tmp/installation_profile' % target)

if not username:
	execute('sed -i "s/\#d-i passwd\/make-user/d-i passwd\/make-user/g" %s/tmp/installation_profile' % target)

for patchfile in ( target + '/tmp/installation_profile' , target+'/tmp/keyboard'):
	execute('/usr/local/bin/patcha %s' % patchfile)
	execute('/usr/local/bin/patcha -f %s %s' % (PATCHA_IN, patchfile))

if install_opsi_client_agent == False:
	lines = []
	with open(target + '/tmp/installation_profile', 'r') as inputpreseed:
		lines = inputpreseed.readlines()
	outlines = []
	keep = True
	for line in lines:
		if "in-target mkdir -p /var/tmp/opsi-linux-client-agent" in line:
			keep = False
		if keep:
			outlines.append(line)
		if "systemctl enable" in line:
			keep = True
	with open(target + '/tmp/installation_profile', 'w') as outputpreseed:
		outputpreseed.writelines(outlines)

######### enable root login on console if password is set
rootPassword = productProperties.get('root_password','')
if rootPassword:
	execute("sed -i '/root-login/s/#//g' %s/tmp/installation_profile" % target)

######## how to install packages

packageDependency = productProperties.get('package_dependency', 'recommends')
if packageDependency != 'recommends':
	execute("sed -i '/install-recommends/s/#//g' %s/tmp/installation_profile" % target)

if install_unattended:
	logger.notice(u"Copy profile file to initrd...")
	copy(target+'/tmp/installation_profile', target+'/tmp/mnt/preseed.cfg')
	logger.notice(u"Here comes the installation profile:")
	execute('cat %s/tmp/mnt/preseed.cfg' % (target))

logger.notice(u"Copy opsi-client-agent file to initrd...")
execute('mkdir -p %s/tmp/mnt/opsitmp/' % (target))
copy(target+'/tmp/sysconf.ini', target+'/tmp/mnt/opsitmp/')
copy(target+'/tmp/keyboard', target+'/tmp/mnt/opsitmp/')
copy(PATCHA_IN, target+'/tmp/mnt/opsitmp/')

execute(f"rm -rf {target}/tmp/mnt/usr/bin/wget")
depot.copy(SCRIPT_PATH + '/installfiles%s/wget' % myarchitectureBits, target+'/tmp/mnt/usr/bin/')
execute(f"chmod +x {target}/tmp/mnt/usr/bin/wget")

execute(f"wget --no-check-certificate -P {target}/tmp/mnt/etc/ssl/certs/ https://{fqdn}:{pckey}@{depotAddress}:4447/ssl/opsi-ca-cert.pem")
execute(f"mkdir -p {target}/tmp/mnt/usr/lib/ssl/")
execute(f"cat {target}/tmp/mnt/etc/ssl/certs/opsi-ca-cert.pem >> {target}/tmp/mnt/usr/lib/ssl/cert.pem")
execute(f"sed -i 's_wget_/usr/bin/wget_g' {target}/tmp/mnt/scripts/casper")


scriptMessageSubject.setMessage("Rebuild patched initrd.... ")
logger.notice(u"Rebuild patched initrd.... ")
try:
	execute("rm %s/tmp/initrd.lz4" % (target))
	execute("rm %s/tmp/initrd.zst" % (target))
except:
	pass

try:
	execute("ls -l %s/tmp/initrd*" % (target))
except:
	pass
execute("cd %s/tmp/mnt ; find . | cpio -o -H newc > %s/tmp/initrd" % (target, target))
try:
	execute("ls -l %s/tmp/initrd*" % (target))
except:
	pass

if installation_method == 'reboot':
	execute('/bin/efibootmgr -c --disk %s --part %s -L installer --loader \\\\BOOTx64.EFI' % (disk.device, grubPartitionNumber))
	setFirstUefiBoot('installer')
	setNextUefiBoot('installer')
	execute('/bin/efibootmgr -v')
	execute(f"cp {target}/tmp/initrd* {grubTarget}")
	execute(f'cp /mnt/iso/casper/initrd.lz {grubTarget}')
	execute(f"cp {target}/tmp/vmlinuz {grubTarget}")
	try:
		execute(f"cp /mnt/iso/EFI/BOOT/BOOTx64.EFI {grubTarget}")
		execute(f"cp /mnt/iso/EFI/BOOT/grubx64.efi {grubTarget}")
	except RuntimeError:
		execute(f"cp /mnt/iso/EFI/boot/bootx64.efi {grubTarget}/BOOTX64.EFI")
		execute(f"cp /mnt/iso/EFI/boot/grubx64.efi {grubTarget}")
	execute(f"cp -r /mnt/iso/boot/ {grubTarget}")
	execute(f"cp {grubTarget}/grub/grub.cfg {grubTarget}/boot/grub/grub.cfg")
	execute(f"cp {target}/tmp/mnt/preseed.cfg {grubTarget}")
else:
	scriptMessageSubject.setMessage("Prepare loading kernel.... ")
	logger.notice(u"Prepare loading kernel ....")
	kernel_file = target+'/tmp/vmlinuz'
	append_line = f'{mint_kernel_append_parameter} locale={locale} keymap={keymap} console-keymaps-at/keymap={keymap} ip=dhcp noprompt boot=casper root=dev/ram0 ramdisk_size=3500000 url=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/iso/{isofilename} initrd=mint/initrd.lz nosplash automatic-ubiquity --'
	inird_file = target+'/tmp/initrd'
	execute("/sbin/kexec --load %s  --append='%s' --initrd=%s" % (kernel_file, append_line, inird_file))

logger.notice(u"handle_setupAfterInstall ....")
handle_setupAfterInstall()

# Set product installation status
productOnClient.setInstallationStatus('unknown')
#productOnClient.setActionProgress('installing')
productOnClient.setActionResult('')
productOnClient.setLastAction('setup')

#scriptMessageSubject.setMessage("Writing log to service.... ")
try:
	# Write log to service
	logger.notice(u"Writing log to service")

	log = codecs.open(LOG_FILE, 'r', 'utf-8', 'replace')
	data = log.read()
	log.close()

	logger.setFileLevel(LOG_ERROR)
	if backend:
		try:
			try:
				ui.getMessageBox().addText(_(u"Writing log to service\n"))
			except:
				pass
			backend.log_write(logType = 'bootimage', data = data, objectId = clientId, append = False)
		except Exception as e:
			logger.error(u"Failed to write log to service: %s" % e)

except Exception as e:
	logger.logException(e)

if installation_method == 'reboot':
	scriptMessageSubject.setMessage("Rebooting machine ")
	logger.notice(u"Rebooting machine")
	reboot()
else:
	scriptMessageSubject.setMessage("Execute new kernel.... ")
	logger.notice(u"Execute new kernel ....")
	execute("/sbin/kexec -e")
